<?php

if (!defined('_PS_VERSION_')) {
    exit;
}

$autoloadPath = dirname(__FILE__) . '/vendor/autoload.php';
if (file_exists($autoloadPath)) {
    require_once $autoloadPath;
}

class kiwiz extends Module
{
    /**
     * @see parent::construct
     * @since 1.0.0
     * @return void
     */
    public function __construct()
    {
        $this->name = 'kiwiz';
        $this->tab = 'administration';
        $this->version = '1.1.1';
        $this->author = 'Kiwiz';
        $this->need_instance = false;
        $this->bootstrap = true;
        $this->module_key = '614900b4dd6fb8a95beb4cad33eab95d';

        parent::__construct();

        $this->displayName = 'Kiwiz';
        $this->description = 'Certification de Factures pour la Loi anti-fraude TVA';
        $this->confirmUninstall = $this->l('Are you sure you want to uninstall this module?');
        $this->ps_versions_compliancy = array('min' => '1.6.1.24', 'max' => _PS_VERSION_);

        // Handle Validate::isHookName issue
        if (version_compare(_PS_VERSION_, '1.7.0.0', '>=')) {
            $this->syncClassName = Kiwiz\Sync::class;
        } else {
            require_once dirname(__FILE__) . '/models/KiwizSync.php';
            $this->syncClassName = KiwizSync::class;
        }
    }

    /**
     * Install the module & needed tables, register on hooks
     *
     * @see parent::install
     * @since 1.0.0
     * @return bool
     */
    public function install()
    {
        // Init configuration vars
        // Authentication vars
        $this->updateConfigurationValue('KIWIZ_API_STATUS', false);
        $this->updateConfigurationValue('KIWIZ_API_USERNAME', '');
        $this->updateConfigurationValue('KIWIZ_API_PASSWD', '');
        $this->updateConfigurationValue('KIWIZ_API_SUB_ID', '');
        if (!$this->getConfigurationValue('KIWIZ_ACTIVATION_DATE')) {
            $this->updateConfigurationValue('KIWIZ_ACTIVATION_DATE', false);
        }
        // Config vars
        $this->updateConfigurationValue('KIWIZ_TEST_MODE', false);
        $this->updateConfigurationValue('KIWIZ_SAVE_LOGS', '');
        $this->updateConfigurationValue('KIWIZ_SEND_LOGS', '');
        $this->updateConfigurationValue('KIWIZ_SEND_LOGS_MAIL', Configuration::get('PS_SHOP_EMAIL'));
        $this->updateConfigurationValue('KIWIZ_QUOTA_THRESHOLD', '50');

        // Create or update secure key for CRON task
        if (!Configuration::getGlobalValue('KIWIZ_SECURE_KEY')) {
            Configuration::updateGlobalValue('KIWIZ_SECURE_KEY', Tools::strtoupper(Tools::passwdGen(32)));
        }

        return parent::install() &&
            $this->createTables() &&
            $this->registerOnHooks();
    }

    /**
     * Register all needed hooks, depending on PrestaShop version
     *
     * @since 1.0.0
     * @return bool
     */
    protected function registerOnHooks()
    {
        $res = $this->registerHook('actionOrderStatusPostUpdate');
        $res = $this->registerHook('actionSetInvoice');
        $res &= $this->registerHook('actionObjectOrderSlipAddAfter');
        $res &= $this->registerHook('actionAdminOrdersListingFieldsModifier');
        $res &= $this->registerHook('actionAdminSlipListingFieldsModifier');
        $res &= $this->registerHook('displayAdminAfterHeader');
        $res &= $this->registerHook('displayBackOfficeOrderActions');
        $res &= $this->registerHook('displayBackOfficeHeader');

        if (version_compare(_PS_VERSION_, '1.7.7.0', '>=')) {
            $res &= $this->registerHook('actionOrderGridDefinitionModifier');
            $res &= $this->registerHook('actionOrderGridPresenterModifier');
            $res &= $this->registerHook('actionCreditSlipGridDefinitionModifier');
            $res &= $this->registerHook('actionCreditSlipGridPresenterModifier');
            $res &= $this->registerHook('actionDispatcherBefore');
        }

        return $res;
    }

    /**
     * Create database tables
     *
     * @since 1.0.0
     * @return bool
     */
    protected function createTables()
    {
        $sql = array();

        $sql[] = 'CREATE TABLE IF NOT EXISTS `'._DB_PREFIX_.'kiwiz_sync` (
            `id_kiwiz_sync` int(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `id_document` int(10) UNSIGNED NOT NULL,
            `document_type` enum("order_invoice", "order_slip") NOT NULL,
            `file_hash` varchar(255) NOT NULL,
            `block_hash` varchar(255) NOT NULL,
            `date_add` datetime NOT NULL,
            PRIMARY KEY (`id_kiwiz_sync`),
            KEY `date_add` (`date_add`),
            UNIQUE KEY `kiwiz_unique` (`id_document`, `document_type`)
        ) ENGINE=' . _MYSQL_ENGINE_ . ' DEFAULT CHARSET=utf8;';

        foreach ($sql as $query) {
            if (Db::getInstance()->execute($query) == false) {
                return false;
            }
        }

        return true;
    }

    /**
     * When an order status change, try to push the invoice (if available) to Kiwiz
     *
     * @since 1.0.0
     * @return void
     */
    public function hookActionOrderStatusPostUpdate($params)
    {
        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        // Check if id_order is defined into hook parameters
        if (empty($params['id_order'])) {
            return;
        }

        // Try to load the order, check object validity & if invoice are available
        $order = new Order($params['id_order']);
        if (!Validate::isLoadedObject($order) || !$order->hasInvoice()) {
            return;
        }
        // Skip order if shop context is not the same as the order
        if ($order->id_shop != $this->context->shop->id) {
            return;
        }

        // Send the invoice to kiwiz
        $this->sendInvoiceToKiwiz($order);
    }

    /**
     * When a new invoice is created, try to push the invoice (if available) to Kiwiz
     *
     * @since 1.0.0
     * @return void
     */
    public function hookActionSetInvoice($params)
    {
        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        // Check if Order is defined into hook parameters
        if (empty($params['Order']) || !($params['Order'] instanceof Order)) {
            return;
        }

        // Try to load the order, check object validity & if invoice are available
        $order = $params['Order'];
        if (!Validate::isLoadedObject($order) || !$order->hasInvoice()) {
            return;
        }
        // Skip order if shop context is not the same as the order
        if ($order->id_shop != $this->context->shop->id) {
            return;
        }

        // Send the invoice to kiwiz
        $this->sendInvoiceToKiwiz($order);
    }

    /**
     * When an order slip is created, try to push it to Kiwiz
     *
     * @since 1.0.0
     * @return void
     */
    public function hookActionObjectOrderSlipAddAfter($params)
    {
        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        // Try to load the order slip, check object validity
        if (empty($params['object']) || !Validate::isLoadedObject($params['object']) || !($params['object'] instanceof OrderSlip)) {
            return;
        }

        // Send the order slip to kiwiz
        $this->sendOrderSlipToKiwiz($params['object']);
    }

    /**
     * Check if the Kiwiz API is ready
     *
     * @since 1.0.0
     * @param int $idShop
     * @return boolean
     */
    public function isApiReady($idShop = null)
    {
        return ($this->getConfigurationValue('KIWIZ_API_STATUS', $idShop) == true);
    }

    /**
     * Check if the current controller is AdminOrders
     *
     * @since 1.0.0
     * @return boolean
     */
    private function isAdminOrdersController()
    {
        if (version_compare(_PS_VERSION_, '1.7.7.0', '>=')) {
            // Try to retrieve current route
            $currentRoute = $this->getRequestParameter('_route');
            return in_array($currentRoute, array('admin_orders_view', 'admin_orders_index'));
        }

        return ($this->context->controller instanceof AdminOrdersController);
    }

    /**
     * Check if the current controller is AdminSlip
     *
     * @since 1.0.0
     * @return boolean
     */
    private function isAdminSlipController()
    {
        if (version_compare(_PS_VERSION_, '1.7.7.0', '>=')) {
            // Try to retrieve current route
            $currentRoute = $this->getRequestParameter('_route');
            return in_array($currentRoute, array('admin_credit_slips_index', 'admin_orders_index'));
        }

        return ($this->context->controller instanceof AdminSlipController);
    }

    /**
     * Retrieve parameter from request component
     *
     * @since 1.0.0
     * @return mixed|null
     */
    private function getRequestParameter($parameter)
    {
        $value = null;
        try {
            $currentRequest = $this->getContainer()->get('request_stack')->getCurrentRequest();
            if (!empty($currentRequest)) {
                $value = $currentRequest->get($parameter);
            }
        } catch (\Exception $e) {
        }

        return $value;
    }

    /**
     * Retrieve the current Order
     *
     * @since 1.0.0
     * @return Order|null
     */
    private function getCurrentOrder()
    {
        if (version_compare(_PS_VERSION_, '1.7.7.0', '>=')) {
            $idOrder = $this->getRequestParameter('orderId');
        } elseif (Tools::getValue('id_order')) {
            $idOrder = (int)Tools::getValue('id_order');
        }
        if (empty($idOrder)) {
            return null;
        }

        return (new Order($idOrder));
    }

    /**
     * Prevent some actions to be triggered (PS 1.7.7+ only)
     *
     * @since 1.0.0
     * @return void
     */
    public function hookActionDispatcherBefore($params)
    {
        // Prevent usage of Kiwiz if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }
        if (version_compare(_PS_VERSION_, '1.7.7.0', '<')) {
            return;
        }

        // Forbidden route
        $forbiddenRoutes = array(
            'admin_order_addresses_edit',
            'admin_orders_add_product',
            'admin_orders_add_cart_rule',
            'admin_orders_remove_cart_rule',
            'admin_orders_change_customer_address',
            'admin_orders_update_invoice_note',
        );
        $withRedirect = array(
            'admin_orders_add_cart_rule',
            'admin_orders_remove_cart_rule',
            'admin_orders_change_customer_address',
            'admin_orders_update_invoice_note',
        );
        $withJGrowl = array(
            'admin_orders_add_product',
        );

        // Try to retrieve current route
        $currentRoute = $this->getRequestParameter('_route');

        // Process bulk synchronization request
        if ($currentRoute == 'admin_orders_search' && Tools::isSubmit('submitBulkForceKiwizSync')) {
            $this->processBulkSynchronization();
        }

        // Unmanaged route
        if (!in_array($currentRoute, $forbiddenRoutes)) {
            return;
        }

        // Try to retrieve current order
        $order = $this->getCurrentOrder();
        if (!$order->hasInvoice()) {
            return;
        }

        // Display an error message
        if (!in_array($currentRoute, $withRedirect)) {
            http_response_code(403);
            if (in_array($currentRoute, $withJGrowl)) {
                header('Content-Type: application/json');
                die(Tools::jsonEncode(
                    array('message' => $this->l('The order cannot be edited'))
                ));
            } else {
                die($this->l('The order cannot be edited'));
            }
        }

        // Redirect to Order page, with error
        Tools::redirectAdmin($this->getContainer()->get('router')->generate(
            'admin_orders_view',
            [
                'orderId' => $order->id,
                'kiwizDeny' => 1,
            ]
        ));
        die;
    }

    /**
     * Add the CSS & JavaScript files you want to be loaded in the BO.
     *
     * @since 1.0.0
     * @return void
     */
    public function hookDisplayBackOfficeHeader()
    {
        if (Tools::getValue('configure') == $this->name) {
            // Module configuration, include assets
            $this->context->controller->addCSS($this->_path . 'views/css/admin-module.css');
        } else {
            // Display an error in case API status is failing and an username is already defined
            if (!$this->getConfigurationValue('KIWIZ_API_STATUS') && $this->getConfigurationValue('KIWIZ_API_USERNAME')) {
                $this->context->controller->errors[] = $this->l('Kiwiz error: API connection is failing, please go to module configuration and fix your credentials');
            }
        }

        // Handle Kiwiz logout
        if (Tools::isSubmit('submitKiwizLogout')) {
            $this->updateConfigurationValue('KIWIZ_API_STATUS', false);
            $this->updateConfigurationValue('KIWIZ_API_USERNAME', '');
            $this->updateConfigurationValue('KIWIZ_API_PASSWD', '');
            $this->updateConfigurationValue('KIWIZ_API_SUB_ID', '');
            $this->updateConfigurationValue('KIWIZ_API_TOKEN', null);
            $this->context->controller->confirmations[] = $this->l('Successfully disconnected.');
        }

        // Do not run the module if it's not actived
        if (empty($this->active)) {
            return;
        }

        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        if ($this->isAdminOrdersController() || $this->isAdminSlipController()) {
            // Order & Order slip listing, add Kiwiz state
            $this->context->controller->addCSS($this->_path . 'views/css/admin-listing.css');
        }
        if ($this->isAdminOrdersController()) {
            // Order page, include assets that hide some actions
            $this->context->controller->addJS($this->_path . 'views/js/admin-order.js');
            $this->context->controller->addCSS($this->_path . 'views/css/admin-order.css');

            // Bulk actions
            // Hide this action into multishop env
            if (Shop::getContext() == Shop::CONTEXT_SHOP) {
                Media::addJsDef(array(
                    'kiwizActions' => array(
                        'bulk' => $this->context->smarty->fetch($this->local_path.'views/templates/admin/order_list/bulk_actions.tpl'),
                    ),
                ));
                // Process bulk synchronization request
                if (Tools::isSubmit('submitBulkForceKiwizSync')) {
                    $this->processBulkSynchronization();
                }
            }

            // Display message to flash (PS 1.7.7+ only)
            if (version_compare(_PS_VERSION_, '1.7.7.0', '>=') && $this->getContainer()->get('session')->get('kiwizMessages')) {
                $kiwizMessages = $this->getContainer()->get('session')->get('kiwizMessages');
                foreach ($kiwizMessages['errors'] as $error) {
                    $this->context->controller->errors[] = $error;
                }
                foreach ($kiwizMessages['confirmations'] as $confirmation) {
                    $this->context->controller->confirmations[] = $confirmation;
                }
                $this->getContainer()->get('session')->remove('kiwizMessages');
            }
        }

        // Try to retrieve the current order, do not proceed if the current order has not any invoice
        $order = $this->getCurrentOrder();

        if (Tools::isSubmit('submitKiwizForceSync')) {
            // Force synchronization for a specific order
            if (Validate::isLoadedObject($order)) {
                $syncResult = $this->processSync(array($order->id));
            } else {
                $syncResult = $this->processSync();
            }

            // Display error message if needed
            if (!empty($syncResult['orders']['errors'])) {
                foreach ($syncResult['orders']['errors'] as $error) {
                    $this->context->controller->errors[] = sprintf($this->l('Kiwiz error: %s'), $error);
                }
            }
            if (!empty($syncResult['order_slip']['errors'])) {
                foreach ($syncResult['order_slip']['errors'] as $error) {
                    $this->context->controller->errors[] = sprintf($this->l('Kiwiz error: %s'), $error);
                }
            }
            // Display confirmation message
            if (empty($syncResult['orders']['errors']) && empty($syncResult['order_slip']['errors'])) {
                $this->context->controller->confirmations[] = sprintf($this->l('Kiwiz: %d new synced document(s)'), ($syncResult['orders']['synced'] + $syncResult['order_slip']['synced']));
            }
        }

        // Do not proceed if the current order has not any invoice
        if (!Validate::isLoadedObject($order) || !$order->hasInvoice()) {
            return;
        }
        // Prevent display of any Kiwiz items if API status is invalid for the current Order shop
        if (!$this->isApiReady($order->id_shop)) {
            return;
        }

        // Hide & prevent specific actions on AdminOrders
        $this->hideAndPreventActions($order);

        // Retrieve state of each document
        $invoice = OrderInvoice::getInvoiceByNumber($order->invoice_number);
        $kiwizDocuments = array(
            'invoice_' . (int)$invoice->id => $this->renderDocumentState('invoice_' . (int)$invoice->id, Kiwiz\Sync::findByObject($order), $order),
        );
        foreach ($order->getOrderSlipsCollection() as $orderSlip) {
            $kiwizDocuments['orderslip_' . (int)$orderSlip->id] = $this->renderDocumentState('orderslip_' . (int)$orderSlip->id, Kiwiz\Sync::findByObject($orderSlip), $orderSlip);
        }

        // Add rendered document states to kiwizDocuments JS variable
        Media::addJsDef(array(
            'kiwizDocuments' => $kiwizDocuments,
        ));
    }

    /**
     * Process bulk action received from Order & OrderSlip controller
     *
     * @since 1.0.0
     * @return void
     */
    protected function processBulkSynchronization()
    {
        // Retrieve selected ID Order
        if (version_compare(_PS_VERSION_, '1.7.7.0', '>=')) {
            $idOrderList = Tools::getValue('order_orders_bulk', array());
        } else {
            $idOrderList = Tools::getValue('orderBox', array());
        }
        if (is_array($idOrderList)) {
            $idOrderList = array_map('intval', $idOrderList);
        } else {
            $idOrderList = array();
        }
        if (empty($idOrderList)) {
            return;
        }
        $kiwizMessages = array(
            'errors' => array(),
            'confirmations' => array(),
        );

        // Process order sync
        $stats = $this->processSync($idOrderList);

        // Display error or confirmation message
        if (!empty($stats['orders']['failed']) || !empty($stats['order_slip']['failed'])) {
            // At least one document failed to sync
            $this->context->controller->errors[] = $kiwizMessages['errors'][] = sprintf($this->l('Kiwiz: %d document(s) failed to sync'), ($stats['orders']['failed'] + $stats['order_slip']['failed']));
        }
        if (!empty($stats['orders']['synced']) || !empty($stats['order_slip']['synced'])) {
            // At least one new document was synced
            $this->context->controller->confirmations[] = $kiwizMessages['confirmations'][] = sprintf($this->l('Kiwiz: %d new synced document(s)'), ($stats['orders']['synced'] + $stats['order_slip']['synced']));
        }
        if (empty($stats['orders']['total']) && empty($stats['order_slip']['total'])) {
            $this->context->controller->confirmations[] = $kiwizMessages['confirmations'][] = $this->l('Kiwiz: nothing to sync');
        }

        // Store message to flash on the next page
        if (version_compare(_PS_VERSION_, '1.7.7.0', '>=')) {
            $this->getContainer()->get('session')->set('kiwizMessages', $kiwizMessages);
        }
    }

    /**
     * Render info for a Kiwiz document
     *
     * @since 1.0.0
     * @param string $docId
     * @param \Kiwiz\Sync|null $document
     * @param \ObjectModel $obj
     * @return array
     */
    protected function renderDocumentState($docId, $document, ObjectModel $obj)
    {
        // Check order date or invoice date
        $shouldBeSynced = true;
        $activationDate = $this->getConfigurationValue('KIWIZ_ACTIVATION_DATE');
        if ($obj instanceof Order) {
            $shouldBeSynced = ($this->getOrderOrInvoiceDate($obj) >= $activationDate);
        } elseif ($obj instanceof OrderSlip) {
            $shouldBeSynced = (strtotime($obj->date_add) >= $activationDate);
        }

        $this->context->smarty->assign(array(
            'kiwizDocId' => $docId,
            'kiwizDocument' => $document,
            'shouldBeSynced' => $shouldBeSynced,
        ));

        return array(
            'state' => $this->context->smarty->fetch($this->local_path.'views/templates/admin/document_state.tpl'),
            'infos' => $this->context->smarty->fetch($this->local_path.'views/templates/admin/document_infos.tpl'),
        );
    }

    /**
     * Hide & prevent specific actions on AdminOrders
     *
     * @since 1.0.0
     * @param \Order $order
     * @return void
     */
    protected function hideAndPreventActions(Order $order)
    {
        if ($this->isAdminOrdersController() && $order) {
            // Order page, include assets that hide some actions
            $this->context->controller->addJS($this->_path . 'views/js/admin-order-hide.js');
            $this->context->controller->addCSS($this->_path . 'views/css/admin-order-hide.css');
        }

        // Display a specific error message in case we deny an action
        if (Tools::getIsset('kiwizDeny')) {
            $this->context->controller->errors[] = $this->l('The order cannot be edited');
        }

        // Prevent some actions from Order page like:
        // Add a product, edit a product
        $actionsToPrevent = array(
            'addProductOnOrder',
            'editProductOnOrder',
        );
        foreach ($actionsToPrevent as $action) {
            if (Tools::getValue('action') == $action) {
                // Display error as JSON because its triggered from an AJAX request
                die(Tools::jsonEncode(array(
                    'result' => false,
                    'error' => $this->l('The order cannot be edited'),
                )));
            }
        }
        // Add a voucher
        // Change shipping or billing address
        // Remove a discount
        $submitActionsToPrevent = array(
            'submitNewVoucher',
            'submitAddressShipping',
            'submitAddressInvoice',
            'submitEditNote',
            'submitDeleteVoucher',
        );
        foreach ($submitActionsToPrevent as $action) {
            if (Tools::isSubmit($action)) {
                // Redirect to Order page and display an error message
                $controller = $this->context->controller;
                Tools::redirectAdmin($controller::$currentIndex . '&kiwizDeny&id_order=' . $order->id . '&vieworder&token=' . $this->context->controller->token);
            }
        }
    }

    /**
     * Add a Kiwiz state block column on Order listing page
     *
     * @since 1.0.0
     * @param array $params
     * @return void
     */
    public function hookActionAdminOrdersListingFieldsModifier($params)
    {
        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        $this->context->smarty->assign(array(
            'kiwizImg' => _MODULE_DIR_  . $this->name . '/views/img/logo.png',
        ));

        $params['fields']['kiwiz'] = array(
            'title' => $this->context->smarty->fetch($this->local_path.'views/templates/admin/order_list/header.tpl'),
            'callback_object' => $this,
            'callback' => 'renderOrderStatusColumn',
            'align' => 'text-center',
            'orderby' => false,
            'search' => false,
        );

        if (isset($params['select'])) {
            $select = trim($params['select'], ',');
            $params['select'] = "'callback' AS `kiwiz`, " . $select;
        }
    }

    /**
     * Add a Kiwiz state block column on Order slip listing page
     *
     * @since 1.0.0
     * @param array $params
     * @return void
     */
    public function hookActionAdminSlipListingFieldsModifier($params)
    {
        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        $this->context->smarty->assign(array(
            'kiwizImg' => _MODULE_DIR_  . $this->name . '/views/img/logo.png',
        ));

        $params['fields']['kiwiz'] = array(
            'title' => $this->context->smarty->fetch($this->local_path.'views/templates/admin/order_list/header.tpl'),
            'callback_object' => $this,
            'callback' => 'renderOrderStatusColumn',
            'align' => 'text-center fixed-width-xs',
            'orderby' => false,
            'search' => false,
            'remove_onclick' => true,
        );

        if (isset($params['select'])) {
            $select = trim($params['select'], ',');
            $params['select'] = "'callback' AS `kiwiz`, " . $select;
        }
    }

    /**
     * Add a Kiwiz state block column on Order listing page (PrestaShop 1.7.7+)
     *
     * @since 1.0.0
     * @param array $params
     * @return void
     */
    public function hookActionOrderGridPresenterModifier($params)
    {
        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        // Retrieve all records
        $params['presented_grid']['data']['records'] = $params['presented_grid']['data']['records']->all();

        // For each record, add the Kiwiz status
        foreach ($params['presented_grid']['data']['records'] as &$row) {
            $row['kiwiz'] = $this->renderOrderStatusColumn(
                '',
                array(
                    'id_order' => $row['id_order'],
                )
            );
        }
    }

    /**
     * Add a Kiwiz column on Order listing page (PrestaShop 1.7.7+)
     *
     * @since 1.0.0
     * @param array $params
     * @return void
     */
    public function hookActionOrderGridDefinitionModifier($params)
    {
        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        // Add new column
        $params['definition']
            ->getColumns()
            ->addAfter(
                'date_add',
                (new Kiwiz\Core\Grid\Column\Type\KiwizColumn('kiwiz'))
                    ->setName('Kiwiz')
                    ->setOptions(
                        [
                            'field' => 'kiwiz',
                        ]
                    )
            );

        // Add bulk action
        if (Shop::getContext() == Shop::CONTEXT_SHOP) {
            $params['definition']
                ->getBulkActions()
                ->add((new PrestaShop\PrestaShop\Core\Grid\Action\Bulk\Type\SubmitBulkAction('bulk_kiwiz_sync'))
                    ->setName($this->l('Force Kiwiz synchronization'))
                    ->setOptions([
                        'submit_route' => 'admin_orders_index',
                        'route_params' => [
                            'submitBulkForceKiwizSync' => '1',
                        ],
                    ])
                );
        }
    }

    /**
     * Add a Kiwiz state block column on Order slip listing page (PrestaShop 1.7.7+)
     *
     * @since 1.0.0
     * @param array $params
     * @return void
     */
    public function hookActionCreditSlipGridPresenterModifier($params)
    {
        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        // Retrieve all records
        $params['presented_grid']['data']['records'] = $params['presented_grid']['data']['records']->all();

        // For each record, add the Kiwiz status
        foreach ($params['presented_grid']['data']['records'] as &$row) {
            $row['kiwiz'] = $this->renderOrderStatusColumn(
                '',
                array(
                    'id_order' => $row['id_order'],
                    'id_order_slip' => $row['id_order_slip'],
                )
            );
        }
    }

    /**
     * Add a Kiwiz column on Order slip listing page (PrestaShop 1.7.7+)
     *
     * @since 1.0.0
     * @param array $params
     * @return void
     */
    public function hookActionCreditSlipGridDefinitionModifier($params)
    {
        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        $params['definition']
            ->getColumns()
            ->addAfter(
                'date_add',
                (new PrestaShop\PrestaShop\Core\Grid\Column\Type\DataColumn('kiwiz'))
                    ->setName('Kiwiz')
                    ->setOptions(
                        [
                            'field' => 'kiwiz',
                        ]
                    )
            );
    }

    /**
     * Render Kiwiz global status for a specific order
     *
     * @since 1.0.0
     * @param string $currentValue
     * @param array $row
     * @return string
     */
    public function renderOrderStatusColumn($currentValue, $row)
    {
        $order = new Order($row['id_order']);
        $orderSlip = null;
        if (!empty($row['id_order_slip'])) {
            $orderSlip = new OrderSlip($row['id_order_slip']);
        }

        // Assign all needed Smarty variables to render Kiwiz state for a specific order
        $this->assignOrderStateVars($order, $orderSlip);

        // Render the Kiwiz state
        return $this->context->smarty->fetch($this->local_path.'views/templates/admin/order_list/sync_state_listing.tpl');
    }

    /**
     * Retrieve newer date for an Order (use Invoice or Order date)
     *
     * @since 1.0.0
     * @param \Order $order
     * @return int
     */
    public function getOrderOrInvoiceDate(Order $order)
    {
        // Check if an invoice exists
        if ($order->hasInvoice()) {
            $orderInvoiceList = $order->getInvoicesCollection();
            $orderInvoice = $orderInvoiceList->getFirst();
            if (Validate::isLoadedObject($orderInvoice) && strtotime($orderInvoice->date_add) > strtotime($order->date_add)) {
                return strtotime($orderInvoice->date_add);
            }
        }

        return strtotime($order->date_add);
    }

    /**
     * Assign all needed Smarty variables to render Kiwiz state for a specific order
     *
     * @since 1.0.0
     * @param \Order $order
     * @param \OrderSlip|null $orderSlip
     * @return void
     */
    protected function assignOrderStateVars(Order $order, OrderSlip $orderSlip = null)
    {
        $nbDocs = 0;
        $shouldBeSynced = true;
        $activationDate = $this->getConfigurationValue('KIWIZ_ACTIVATION_DATE', $order->id_shop);

        // Prevent display of any Kiwiz items if API status is invalid for the current Order shop
        if (!$this->isApiReady($order->id_shop)) {
            $this->context->smarty->assign(array(
                'shouldBeSynced' => false,
                'ordersToSync' => 0,
                'orderSlipToSync' => 0,
                'nbDocs' => 0,
                'nbDocsToSync' => 0,
                'displayForceSyncButton' => false,
            ));
            return;
        }

        // Check order date or invoice date
        if ($this->getOrderOrInvoiceDate($order) < $activationDate) {
            // Do not sync
            $shouldBeSynced = false;
        }

        // Look for orders
        $ordersToSync = array();
        if (empty($orderSlip) && $order->hasInvoice()) {
            $ordersToSync = Kiwiz\Sync::getOrdersToSync(array($order->id));
            $nbDocs++;
        }

        // Look for order slip
        $idOrderSlipList = array();
        if (!empty($orderSlip)) {
            // Force the order slip list to the current one
            $idOrderSlipList[] = (int)$orderSlip->id;
            $nbDocs++;
            $this->context->smarty->assign(array(
                'kiwizDocument' => Kiwiz\Sync::findByObject($orderSlip),
            ));
        } else {
            // Retrieve all order slip for the current order
            foreach ($order->getOrderSlipsCollection() as $orderSlip) {
                $idOrderSlipList[] = (int)$orderSlip->id;
                $nbDocs++;
            }
        }

        $orderSlipToSync = array();
        if (count($idOrderSlipList)) {
            $orderSlipToSync = Kiwiz\Sync::getOrderSlipToSync($idOrderSlipList);
        }
        $nbDocsToSync = count($ordersToSync) + count($orderSlipToSync);

        $this->context->smarty->assign(array(
            'shouldBeSynced' => $shouldBeSynced,
            'ordersToSync' => $ordersToSync,
            'orderSlipToSync' => $orderSlipToSync,
            'nbDocs' => $nbDocs,
            'nbDocsToSync' => $nbDocsToSync,
            'displayForceSyncButton' => ($shouldBeSynced && $nbDocsToSync) && ($order->id_shop == $this->context->shop->id),
        ));
    }

    /**
     * Add a Kiwiz state block on Order page
     *
     * @since 1.0.0
     * @param array $params
     * @return string
     */
    public function hookDisplayBackOfficeOrderActions($params)
    {
        $order = new Order($params['id_order']);

        // Prevent display of any Kiwiz items if API status is invalid for the current Order shop
        if (!$this->isApiReady($order->id_shop)) {
            return;
        }

        // Assign all needed Smarty variables to render Kiwiz state for a specific order
        $this->assignOrderStateVars($order);

        // Render the state
        return $this->context->smarty->fetch($this->local_path.'views/templates/admin/sync_state.tpl');
    }

    /**
     * Add a Kiwiz state block on Order page (PrestaShop 1.6 & PrestaShop 1.7.7+)
     *
     * @since 1.0.0
     * @param array $params
     * @return string
     */
    public function hookDisplayAdminAfterHeader($params)
    {
        // Prevent display of any Kiwiz items if API status is invalid
        if (!$this->isApiReady()) {
            return;
        }

        $order = $this->getCurrentOrder();
        // Do not trigger the hook in some case:
        // Order is NULL
        // PrestaShop between 1.7.0 & 1.7.6
        if ($order === null || (version_compare(_PS_VERSION_, '1.7.0.0', '>=') && version_compare(_PS_VERSION_, '1.7.7.0', '<'))) {
            return;
        }
        // Check that we really are on a order page (not address editin, for example)
        if (version_compare(_PS_VERSION_, '1.7.7.0', '>=') && !$this->isAdminOrdersController()) {
            return;
        }

        // Assign id_order
        $params['id_order'] = (int)$order->id;
        // Set the selector where we want to move the sync block
        $moveToSelector = '.col-lg-7 .panel div.well';
        if (version_compare(_PS_VERSION_, '1.7.7.0', '>=')) {
            $moveToSelector = '.order-actions';
        }
        $this->context->smarty->assign(array(
            'moveSyncBlockTo' => $moveToSelector,
        ));

        return $this->hookDisplayBackOfficeOrderActions($params);
    }

    /**
     * Create the form that will be displayed in the configuration of your module
     *
     * @since 1.0.0
     * @param string $tabName
     * @return string
     */
    protected function renderForm($tabName)
    {
        // Create the HelperForm instance
        $helper = new HelperForm();
        $helper->show_toolbar = false;
        $helper->table = $this->table;
        $helper->module = $this;
        $helper->default_form_language = $this->context->language->id;
        $helper->allow_employee_form_lang = Configuration::get('PS_BO_ALLOW_EMPLOYEE_FORM_LANG', 0);
        $helper->identifier = $this->identifier;
        // Define custom submit action name so it can be identified into postProcess method
        $helper->submit_action = 'submitKiwiz' . $tabName;
        $helper->currentIndex = $this->context->link->getAdminLink('AdminModules', false) . '&configure=' . $this->name . '&tab_module=' . $this->tab . '&module_name=' . $this->name;
        $helper->token = Tools::getAdminTokenLite('AdminModules');

        // Assign common vars
        $helper->tpl_vars = array(
            'fields_value' => $this->getConfigFormValues($tabName),
            'languages' => $this->context->controller->getLanguages(),
            'id_language' => $this->context->language->id,
        );

        // Configuration tab
        if ($tabName == 'config') {
            try {
                // Inject some external data
                $quota = $this->getApiClient()->getQuota();
                // Define some threshold to colorize the progress bar
                $quotaCssClass = 'info';
                if ($quota->consumption > 90) {
                    $quotaCssClass = 'danger';
                } elseif ($quota->consumption > 70) {
                    $quotaCssClass = 'warning';
                }

                // Check quota usage and send an alert in case threshold has been reached
                $this->checkQuotaConsumption($quota);
            } catch (Exception $e) {
                $this->updateConfigurationValue('KIWIZ_API_TOKEN', null);
                $this->updateConfigurationValue('KIWIZ_API_STATUS', false);
                $this->context->controller->errors[] = sprintf($this->l('Kiwiz error: API connection is failing (code %d). Please try again later or contact our support.'), $e->getCode());
                return '';
            }

            $this->context->smarty->assign(array(
                // Quota information
                'kiwizQuota' => $quota,
                'kiwizQuotaProgressClass' => $quotaCssClass,
                // CRON information
                'kiwizCronUrl' => $this->context->link->getModuleLink(
                    'kiwiz',
                    'cron',
                    array(
                        'kiwiz_sk' => Configuration::getGlobalValue('KIWIZ_SECURE_KEY'),
                    )
                ),
                'kiwizCronLastCall' => ($this->getConfigurationValue('KIWIZ_CRON_LAST_CALL') ? Tools::displayDate(date('Y-m-d H:i:s', $this->getConfigurationValue('KIWIZ_CRON_LAST_CALL')), null, true) : $this->l('Never')),
            ));
        }

        // Use the helper to generate our form to HTML
        return $helper->generateForm(array($this->getConfigForm($tabName)));
    }

    /**
     * Create the structure of your form
     *
     * @since 1.0.0
     * @param string $tabName
     * @return array
     */
    protected function getConfigForm($tabName)
    {
        $kiwizApiStatus = $this->isApiReady();
        $formGroupClass = ($kiwizApiStatus === false ? ($this->getConfigurationValue('KIWIZ_API_USERNAME') ? 'has-error' : '') : 'has-success');

        if ($tabName == 'auth') {
            return array(
                'form' => array(
                    'legend' => array(
                        'title' => $this->l('Kiwiz authentication'),
                        'icon' => 'icon-key',
                    ),
                    'input' => array(
                        array(
                            'type' => 'hidden',
                            'name' => 'tabName',
                        ),
                        array(
                            'form_group_class' => $formGroupClass,
                            'required' => true,
                            'type' => 'text',
                            'prefix' => '<i class="icon icon-envelope"></i>',
                            'desc' => $this->l('Enter the email address associated to your Kiwiz account'),
                            'name' => 'KIWIZ_API_USERNAME',
                            'label' => $this->l('Email address'),
                            'placeholder' => 'username@domain.com',
                        ),
                        array(
                            'form_group_class' => $formGroupClass,
                            'required' => true,
                            'type' => 'password',
                            'prefix' => '<i class="icon icon-key"></i>',
                            'desc' => $this->l('Enter the password associated to your Kiwiz account'),
                            'name' => 'KIWIZ_API_PASSWD',
                            'label' => $this->l('Password'),
                            'placeholder' => ''
                        ),
                        array(
                            'form_group_class' => $formGroupClass,
                            'required' => true,
                            'type' => 'text',
                            'prefix' => '<i class="icon icon-info-circle"></i>',
                            'desc' => $this->l('Enter the subscription ID associated to your Kiwiz account'),
                            'name' => 'KIWIZ_API_SUB_ID',
                            'label' => $this->l('Subscription ID'),
                            'placeholder' => '1234567890000123456',
                        ),
                    ),
                    'submit' => array(
                        'title' => $this->l('Save'),
                    ),
                ),
            );
        } elseif ($tabName == 'config') {
            return array(
                'form' => array(
                    'legend' => array(
                        'title' => $this->l('Kiwiz information & configuration'),
                        'icon' => 'icon-cog',
                    ),
                    'input' => array(
                        array(
                            'type' => 'hidden',
                            'name' => 'tabName',
                        ),
                        array(
                            'type' => 'html',
                            'html_content' => '<h2>'. $this->l('Quota information') .'</h2>',
                            'name' => '',
                            'label' => '',
                            'col' => 12,
                        ),
                        array(
                            'type' => 'quotaInformation',
                            'name' => '',
                            'label' => $this->l('Current consumption'),
                            'col' => 5,
                        ),
                        array(
                            'type' => 'html',
                            'html_content' => '<h2>'. $this->l('Test mode') .'</h2>',
                            'name' => '',
                            'label' => '',
                            'col' => 12,
                        ),
                        array(
                            'type' => 'switch',
                            'label' => $this->l('Enable the test mode'),
                            'name' => 'KIWIZ_TEST_MODE',
                            'is_bool' => true,
                            'desc' => $this->l('Transactions will still be saved but not subject to calculation'),
                            'values' => array(
                                array(
                                    'id' => 'active_on',
                                    'value' => true,
                                    'label' => $this->l('Enabled'),
                                ),
                                array(
                                    'id' => 'active_off',
                                    'value' => false,
                                    'label' => $this->l('Disabled'),
                                )
                            ),
                        ),
                        array(
                            'type' => 'html',
                            'html_content' => '<h2>'. $this->l('Logs configuration') .'</h2>',
                            'name' => '',
                            'label' => '',
                            'col' => 12,
                        ),
                        array(
                            'type' => 'switch',
                            'label' => $this->l('Enable API errors logging'),
                            'name' => 'KIWIZ_SAVE_LOGS',
                            'is_bool' => true,
                            'desc' => '',
                            'values' => array(
                                array(
                                    'id' => 'active_on',
                                    'value' => true,
                                    'label' => $this->l('Enabled'),
                                ),
                                array(
                                    'id' => 'active_off',
                                    'value' => false,
                                    'label' => $this->l('Disabled'),
                                )
                            ),
                        ),
                        array(
                            'type' => 'switch',
                            'label' => $this->l('Send logs by mail'),
                            'name' => 'KIWIZ_SEND_LOGS',
                            'is_bool' => true,
                            'values' => array(
                                array(
                                    'id' => 'active_on',
                                    'value' => true,
                                    'label' => $this->l('Enabled'),
                                ),
                                array(
                                    'id' => 'active_off',
                                    'value' => false,
                                    'label' => $this->l('Disabled'),
                                )
                            ),
                        ),
                        array(
                            'type' => 'text',
                            'label' => $this->l('Email addresses that will receive the alerts'),
                            'name' => 'KIWIZ_SEND_LOGS_MAIL',
                            'desc' => $this->l('Separate each email address using a comma'),
                            'col' => 4,
                        ),
                        array(
                            'type' => 'select',
                            'label' => $this->l('Quota threshold alert'),
                            'name' => 'KIWIZ_QUOTA_THRESHOLD',
                            'desc' => $this->l('Send a daily email alert in case quota consumption is equal or higher to this threshold'),
                            'options' => array(
                                'query' => $this->createQuotaThresholdOptions(),
                                'id' => 'value',
                                'name' => 'label',
                            ),
                        ),
                        array(
                            'type' => 'html',
                            'html_content' => '<h2>'. $this->l('CRON') .'</h2>',
                            'name' => '',
                            'label' => '',
                            'col' => 12,
                        ),
                        array(
                            'type' => 'cronInformation',
                            'name' => '',
                            'label' => '',
                            'col' => 12,
                        ),
                    ),
                    'submit' => array(
                        'title' => $this->l('Save'),
                    ),
                ),
            );
        }
    }

    /**
     * Create options for <select> quota threshold item
     *
     * @since 1.0.0
     * @return array
     */
    protected function createQuotaThresholdOptions()
    {
        $values = [];

        for ($i = 5; $i <= 100; $i = $i+5) {
            $values[] = array(
                'value' => $i,
                'label' => $i . ' %',
            );
        }

        return $values;
    }

    /**
     * Set values for the inputs
     *
     * @since 1.0.0
     * @param string $tabName
     * @return array
     */
    protected function getConfigFormValues($tabName)
    {
        if ($tabName == 'auth') {
            return array(
                'KIWIZ_API_USERNAME' => $this->getConfigurationValue('KIWIZ_API_USERNAME'),
                'KIWIZ_API_PASSWD' => $this->getConfigurationValue('KIWIZ_API_PASSWD'),
                'KIWIZ_API_SUB_ID' => $this->getConfigurationValue('KIWIZ_API_SUB_ID'),
                'tabName' => $tabName,
            );
        } elseif ($tabName == 'config') {
            return array(
                'KIWIZ_TEST_MODE' => (bool)$this->getConfigurationValue('KIWIZ_TEST_MODE'),
                'KIWIZ_SAVE_LOGS' => (bool)$this->getConfigurationValue('KIWIZ_SAVE_LOGS'),
                'KIWIZ_SEND_LOGS' => (bool)$this->getConfigurationValue('KIWIZ_SEND_LOGS'),
                'KIWIZ_SEND_LOGS_MAIL' => $this->getConfigurationValue('KIWIZ_SEND_LOGS_MAIL'),
                'KIWIZ_QUOTA_THRESHOLD' => (int)$this->getConfigurationValue('KIWIZ_QUOTA_THRESHOLD'),
                'tabName' => $tabName,
            );
        }
    }

    /**
     * Save form data
     *
     * @since 1.0.0
     * @return void
     */
    protected function postProcess()
    {
        // Do not process if we aren't into a submit configuration case
        $tabName = Tools::getValue('tabName');
        if (!$tabName || !in_array($tabName, ['auth', 'config']) || !Tools::isSubmit('submitKiwiz' . $tabName)) {
            return;
        }

        // Save new values to configuration table
        $formValues = $this->getConfigFormValues($tabName);
        foreach (array_keys($formValues) as $key) {
            $value = Tools::getValue($key);
            // Ignore this key
            if ($key == 'tabName') {
                continue;
            }
            if ($key == 'KIWIZ_API_PASSWD') {
                // Force reset of API status in case auth form is submitted
                if ($this->getConfigurationValue('KIWIZ_API_STATUS')) {
                    $this->updateConfigurationValue('KIWIZ_API_STATUS', false);
                }
                // Encrypt password
                $value = $this->encryptData($value);
            }
            // List of email address for alert purposes
            if ($key == 'KIWIZ_SEND_LOGS_MAIL') {
                $emails = array();
                foreach (explode(',', $value) as $email) {
                    $email = trim($email);
                    if (Validate::isEmail($email)) {
                        $emails[] = trim($email);
                    }
                }
                $value = implode(',', $emails);
            }
            $this->updateConfigurationValue($key, $value);
        }

        if ($tabName == 'auth') {
            // Try to generate a token & save it
            try {
                // Ask Kiwiz API for a token
                $this->generateAndBackupApiToken();

                // Display a confirmation message
                $this->context->controller->confirmations[] = $this->l('Credentials are OK. You can start using Kiwiz now!');
            } catch (Exception $e) {
                // Invalidate current API status
                $this->updateConfigurationValue('KIWIZ_API_STATUS', false);

                // Display an error message + API error
                $this->context->controller->errors[] = $this->l('Unable to validate theses credentials, please check them or contact Kiwiz support.');
                $this->context->controller->errors[] = $e->getMessage();
            }
        } else {
            // Display a confirmation message
            $this->context->controller->confirmations[] = $this->l('Configuration successfully updated.');
        }
    }

    /**
     * Ask & backup API token for future Kiwiz API calls
     *
     * @since 1.0.0
     * @param bool $fromRetry
     * @return string
     */
    public function generateAndBackupApiToken($fromRetry = false)
    {
        // Ask Kiwiz API for a token
        $token = $this->getApiClient()->generateToken($this->getConfigurationValue('KIWIZ_API_USERNAME'), $this->decryptData($this->getConfigurationValue('KIWIZ_API_PASSWD')), $this->getConfigurationValue('KIWIZ_API_SUB_ID'));

        // Backup the token
        $this->updateConfigurationValue('KIWIZ_API_STATUS', true);
        $this->updateConfigurationValue('KIWIZ_API_TOKEN', $token);

        // Try to retrieve quota, when auth form is filled
        // Allow us to check subscription ID
        if (!$fromRetry) {
            $quota = $this->getApiClient()->getQuota();
        }

        // Set activation date (if not already defined)
        if (!$this->getConfigurationValue('KIWIZ_ACTIVATION_DATE')) {
            $this->updateConfigurationValue('KIWIZ_ACTIVATION_DATE', time());
        }

        return $token;
    }

    /**
     * Retrieve API client wrapper
     *
     * @since 1.0.0
     * @return \Kiwiz\Api\Client
     */
    private function getApiClient()
    {
        static $client = null;

        if ($client === null) {
            // Retrieve token, if exists & API status is valid
            $token = null;

            if ($this->getConfigurationValue('KIWIZ_API_STATUS') !== false && $this->getConfigurationValue('KIWIZ_API_TOKEN') !== null) {
                $token = $this->getConfigurationValue('KIWIZ_API_TOKEN');
            }
            // Retrieve test mode value
            $testMode = (bool)$this->getConfigurationValue('KIWIZ_TEST_MODE');

            // Create the client
            $client = new Kiwiz\Api\Client($token, $testMode, $this->version);
        }

        return $client;
    }

    /**
     * Load the configuration form
     *
     * @since 1.0.0
     * @return string
     */
    public function getContent()
    {
        // Add alert in case context != shop
        if (Shop::getContext() != Shop::CONTEXT_SHOP) {
            $this->context->controller->errors[] = $this->l('You must select a specific shop in order to continue.');
            return;
        }

        // Check if CURL is available
        if (!extension_loaded('curl') || !function_exists('curl_init')) {
            $this->context->controller->errors[] = $this->l('PHP curl extension is missing, please ask your technical contact to enable it before continue.');
            // Save API status
            $this->updateConfigurationValue('KIWIZ_API_STATUS', false);
            return;
        }

        // If values have been submitted in the form, process.
        $this->postProcess();

        // Select the right configuration form to display regarding API status
        $tabName = 'config';
        if (!$this->isApiReady()) {
            $tabName = 'auth';
        }

        // Kiwiz user information
        if ($tabName == 'config') {
            $this->context->smarty->assign(array(
                'kiwizEmail' => $this->getConfigurationValue('KIWIZ_API_USERNAME'),
                'kiwizSubscriptionId' => $this->getConfigurationValue('KIWIZ_API_SUB_ID'),
            ));
        }

        return $this->context->smarty->fetch($this->local_path.'views/templates/admin/header.tpl') . $this->renderForm($tabName);
    }

    /**
     * Retrieve cipher depending on PS version
     *
     * @since 1.0.0
     * @return \Rijndael|\PhpEncryption
     */
    private function getCipher()
    {
        static $cipher = null;

        if ($cipher === null) {
            if (version_compare(_PS_VERSION_, '1.7.0.0', '>=')) {
                $cipher = new PhpEncryption(_NEW_COOKIE_KEY_);
            } else {
                // Get key depending on shop configuration
                if (!Configuration::get('PS_CIPHER_ALGORITHM') || !defined('_RIJNDAEL_KEY_')) {
                    $key = _COOKIE_KEY_;
                    $iv = _COOKIE_IV_;
                } else {
                    $key = _RIJNDAEL_KEY_;
                    $iv = _RIJNDAEL_IV_;
                }
                $cipher = new Rijndael($key, $iv);
            }
        }

        return $cipher;
    }

    /**
     * Decrypt a specific data
     *
     * @since 1.0.0
     * @param string $data
     * @return mixed
     */
    private function decryptData($data)
    {
        return json_decode($this->getCipher()->decrypt($data));
    }

    /**
     * Encrypt a specific data
     *
     * @since 1.0.0
     * @param mixed $data
     * @return string
     */
    private function encryptData($data)
    {
        return $this->getCipher()->encrypt(json_encode($data));
    }

    /**
     * Try to send the invoice to Kiwiz & add the document to our database
     *
     * @since 1.0.0
     * @param \Order $order
     * @return bool
     */
    public function sendInvoiceToKiwiz(Order $order)
    {
        // Skip order if shop context is not the same as the order
        if ($order->id_shop != $this->context->shop->id) {
            $this->lastApiError = new Exception(sprintf($this->l('Order %d is not associated to the shop %d'), $order->id, $this->context->shop->id));
            return false;
        }

        try {
            // Try to retrieve an existing document first
            $kiwizDocument = Kiwiz\Sync::findByObject($order);
            // Document already exists
            if (Validate::isLoadedObject($kiwizDocument)) {
                return true;
            }

            // Add the document to our database
            $kiwizResult = $this->getApiClient()->saveInvoice($order);
            $kiwizDocument = $this->syncClassName::findOrCreateInvoice($order, $kiwizResult);
            $this->lastApiError = null;

            return true;
        } catch (Exception $e) {
            // Backup last error
            $this->lastApiError = $e;
            // Do not throw an exception here, we'll retry using CRON
            return false;
        }
    }

    /**
     * Try to send the order slip to Kiwiz & add the document to our database
     *
     * @since 1.0.0
     * @param \OrderSlip $orderSlip
     * @return bool
     */
    public function sendOrderSlipToKiwiz(OrderSlip $orderSlip)
    {
        $order = new Order($orderSlip->id_order);

//        dump($order);
//        die;
        // Try to load the order, check object validity
        if (!Validate::isLoadedObject($order)) {
            $this->lastApiError = new Exception(sprintf($this->l('Order %d not found'), $orderSlip->id_order));
            return false;
        }
        // Skip order if shop context is not the same as the order
        if ($order->id_shop != $this->context->shop->id) {
            $this->lastApiError = new Exception(sprintf($this->l('Order %d is not associated to the shop %d'), $order->id, $this->context->shop->id));
            return false;
        }

        try {
            // Try to retrieve an existing document first
            $kiwizDocument = Kiwiz\Sync::findByObject($orderSlip);
            // Document already exists
            if (Validate::isLoadedObject($kiwizDocument)) {
                return true;
            }

            // Add the document to our database
            // hmzeh

            $kiwizResult = $this->getApiClient()->saveCreditMemo($orderSlip);
            $kiwizDocument = $this->syncClassName::findOrCreateCreditMemo($orderSlip, $kiwizResult);
            $this->lastApiError = null;

            return true;
        } catch (Exception $e) {
            // Backup last error
            $this->lastApiError = $e;
            // Do not throw an exception here, we'll retry using CRON
            return false;
        }
    }

    /**
     * Run sync tasks related to the CRON or bulk actions
     *
     * @since 1.0.0
     * @param array $idOrderList
     * @param bool $fromCron
     * @return array
     */
    public function processSync(array $idOrderList = array(), $fromCron = false)
    {
        $idOrderSlipList = array();

        // Stats
        $stats = array(
            'orders' => array(
                'synced' => 0,
                'failed' => 0,
            ),
            'order_slip' => array(
                'synced' => 0,
                'failed' => 0,
            ),
        );

        // Prevent usage of Kiwiz if API status is invalid
        if (!$this->isApiReady()) {
            return $stats;
        }

        // Look for orders
        $ordersToSync = Kiwiz\Sync::getOrdersToSync($idOrderList);
        foreach ($ordersToSync as $idOrder) {
            // Try to load the order, check object validity & if invoice are available
            $order = new Order($idOrder);
            if (!Validate::isLoadedObject($order) || !$order->hasInvoice()) {
                $stats['orders']['failed']++;
                continue;
            }
            // Skip order if shop context is not the same as the order
            if ($order->id_shop != $this->context->shop->id) {
                continue;
            }

            // Try to send the document to Kiwiz
            if ($this->sendInvoiceToKiwiz($order)) {
                $stats['orders']['synced']++;
            } else {
                $stats['orders']['failed']++;
                // Save error message for debug usage
                if (!isset($stats['orders']['errors'])) {
                    $stats['orders']['errors'] = array();
                }
                if (isset($this->lastApiError)) {
                    $stats['orders']['errors'][(int)$order->id] = $this->lastApiError->getMessage();
                    $this->lastApiError = null;
                }
            }

            if (!empty($idOrderList)) {
                // Retrieve all order slip for the current order
                foreach ($order->getOrderSlipsCollection() as $orderSlip) {
                    $idOrderSlipList[] = (int)$orderSlip->id;
                }
            }
        }
        $stats['orders']['total'] = count($ordersToSync);

        // Look for order slip to sync, but order is already synced
        foreach ($idOrderList as $idOrder) {
            // Try to load the order, check object validity & if invoice are available
            $order = new Order($idOrder);
            if (!Validate::isLoadedObject($order) || !$order->hasInvoice()) {
                continue;
            }
            // Skip order if shop context is not the same as the order
            if ($order->id_shop != $this->context->shop->id) {
                continue;
            }

            // Retrieve all order slip for the current order
            foreach ($order->getOrderSlipsCollection() as $orderSlip) {
                $idOrderSlipList[] = (int)$orderSlip->id;
            }
        }
        $idOrderSlipList = array_unique($idOrderSlipList);

        // Look for order slip, if order not submitted
        if (!empty($idOrderList) && empty($idOrderSlipList)) {
            $orderSlipToSync = array();
        } else {
            $orderSlipToSync = Kiwiz\Sync::getOrderSlipToSync($idOrderSlipList);
        }

        foreach ($orderSlipToSync as $idOrderSlip) {
            // Try to load the order slip, check object validity
            $orderSlip = new OrderSlip($idOrderSlip);
            if (!Validate::isLoadedObject($orderSlip)) {
                $stats['order_slip']['failed']++;
                continue;
            }

            // Try to send the document to Kiwiz
            if ($this->sendOrderSlipToKiwiz($orderSlip)) {
                $stats['order_slip']['synced']++;
            } else {
                $stats['order_slip']['failed']++;
                // Save error message for debug usage
                if (!isset($stats['order_slip']['errors'])) {
                    $stats['order_slip']['errors'] = array();
                }
                if (isset($this->lastApiError)) {
                    $stats['order_slip']['errors'][(int)$orderSlip->id] = $this->lastApiError->getMessage();
                    $this->lastApiError = null;
                }
            }
        }
        $stats['order_slip']['total'] = count($orderSlipToSync);

        // Save last cron task call
        if ($fromCron) {
            $this->updateConfigurationValue('KIWIZ_CRON_LAST_CALL', time());
        }

        // Check quota usage and send an alert in case threshold has been reached
        $this->checkQuotaConsumption();

        // Expose some stats
        return $stats;
    }

    /**
     * Method called from PDF override
     *
     * @since 1.0.0
     * @param \PDF $pdfObject
     * @param boolean $display
     * @return mixed
     */
    public function renderPdf(PDF $pdfObject, $display = true)
    {
        $kiwizPdfRenderer = new Kiwiz\PdfRenderer($this->getApiClient(), $pdfObject, $display);

        return $kiwizPdfRenderer->render();
    }

    /**
     * Check quota usage and send an alert in case threshold has been reached
     *
     * @since 1.0.0
     * @param \stdClass $quota
     * @return void
     */
    protected function checkQuotaConsumption(\stdClass $quota = null)
    {
        $lastCall = $this->getConfigurationValue('KIWIZ_QUOTA_LAST_CALL');
        if (!empty($lastCall) && ($lastCall + 86400) > time()) {
            // Check one time, each day
            return;
        }

        // Retrieve current quota, if not directly provided
        if ($quota === null) {
            $quota = $this->getApiClient()->getQuota();
        }

        // Trigger alert in case quota threshold is triggered
        if ($quota->consumption >= $this->getConfigurationValue('KIWIZ_QUOTA_THRESHOLD')) {
            // Send the alert to the shop owner
            foreach (explode(',', $this->getConfigurationValue('KIWIZ_SEND_LOGS_MAIL')) as $email) {
                // Validate email address
                if (!Validate::isEmail($email)) {
                    continue;
                }

                Mail::Send(
                    (int)$this->context->language->id,
                    'quota_alert',
                    $this->l('Kiwiz: quota alert'),
                    array(
                        '{quota_threshold}' => (int)$this->getConfigurationValue('KIWIZ_QUOTA_THRESHOLD')
                    ),
                    $email,
                    null,
                    null,
                    null,
                    null,
                    null,
                    dirname(__FILE__) . '/mails/'
                );
            }
        }

        // Update last quota check timestamp
        $this->updateConfigurationValue('KIWIZ_QUOTA_LAST_CALL', time());
    }

    /**
     * Send an alert to shop owner in case an Exception is triggered
     *
     * @since 1.0.0
     * @param \Exception $exception
     * @return void
     */
    public function sendAlert(\Exception $exception)
    {
        // Send the alert by mail
        foreach (explode(',', $this->getConfigurationValue('KIWIZ_SEND_LOGS_MAIL')) as $email) {
            // Validate email address
            if (!Validate::isEmail($email)) {
                continue;
            }

            Mail::Send(
                (int)Configuration::get('PS_LANG_DEFAULT'),
                'log_alert',
                $this->l('Kiwiz error'),
                array(
                    '{error_message}' => $exception->getMessage(),
                    '{error_code}' => $exception->getCode(),
                ),
                $email,
                null,
                null,
                null,
                null,
                null,
                dirname(__FILE__) . '/mails/'
            );
        }
    }

    /**
     * Get configuration value & force id_shop to be defined
     *
     * @since 1.0.0
     * @param string $key
     * @param int $idShop
     * @return string
     */
    protected function getConfigurationValue($key, $idShop = null)
    {
        if ($idShop === null) {
            $idShop = (int)$this->context->shop->id;
        }

        return Configuration::get($key, null, null, (int)$idShop);
    }

    /**
     * Update configuration value & force id_shop to be defined
     *
     * @since 1.0.0
     * @param string $key
     * @param mixed $values
     * @return bool
     */
    protected function updateConfigurationValue($key, $values)
    {
        return Configuration::updateValue($key, $values, false, null, $this->context->shop->id);
    }
}
