<?php

namespace Kiwiz\Api;

use \Order;
use \Module;
use \OrderSlip;
use Kiwiz\Sync;
use Kiwiz\ObjectMapper;
use Kiwiz\PdfRetriever;
use GuzzleHttp\Psr7\Request;
use GuzzleHttp\Post\PostFile;
use GuzzleHttp\Client as GuzzleClient;
use Kiwiz\Exceptions\ApiErrorException;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Event\AbstractTransferEvent;
use Kiwiz\Exceptions\MalformedJsonException;
use GuzzleHttp\Subscriber\Retry\RetrySubscriber;
if (!defined('_PS_VERSION_')) { exit; }
class Client
{
    /**
     * API endpoint URL
     *
     * @var string
     */
    const API_URL = 'https://api.kiwiz.io';

    /**
     * Create Guzzle client for Kiwiz API & init others components
     *
     * @param string|null $token
     * @param bool|null $testMode
     * @param string $moduleVersion
     * @return void
     */
    public function __construct($token = null, $testMode = false, $moduleVersion = '')
    {
        $this->client = new GuzzleClient([
            'base_url' => self::API_URL,
            // Define specific headers
            'defaults' => [
                'headers' => [
                    'Accept' => 'application/json',
                ],
            ],
        ]);
        // Test mode ?
        $this->testMode = $testMode;
        // Module version
        $this->moduleVersion = $moduleVersion;

        // Define the current token for all our requests
        if (!empty($token)) {
            $this->client->setDefaultOption('headers/Authorization', $token);
        }

        // Retry in case of 401 responses (token has expired)
        // Do not retry if error is related to *subscription id*
        $retry = new RetrySubscriber([
            'filter' => function ($retries, AbstractTransferEvent $event) {
                $resource = $event->getRequest()->getResource();
                // Check for a status code, can be null into specific case like timeout, dns issues...
                $code = $event->getResponse() ? $event->getResponse()->getStatusCode() : null;
                $message = $event->getResponse() ? (string)$event->getResponse() : '';

                if ($code == 401 && !preg_match('#.*subscription\s+id.*#i', $message)) {
                    // Retrieve a new token via module instance
                    $token = Module::getInstanceByName('kiwiz')->generateAndBackupApiToken(true);

                    // Update Authorization header of the client + the current request for the next tries
                    $this->client->setDefaultOption('headers/Authorization', $token);
                    $event->getRequest()->setHeader('Authorization', $token);
                }

                return ($code == 401 && !preg_match('#.*subscription\s+id.*#i', $message));
            },
            'max' => 3,
        ]);

        // Attach our RetrySubscriber instance to the Guzzle client
        $this->client->getEmitter()->attach($retry);

        // Init object mapper & PDF
        $this->objectMapper = new ObjectMapper();
        $this->pdfRetriever = new PdfRetriever();
    }

    /**
     * Retrieve default parameters to send to Kiwiz API
     *
     * @return array
     */
    protected function getDefaultParameters()
    {
        return [
            'platform' => 'PrestaShop',
            'version' => 'prestashop:' . _PS_VERSION_ . '|kiwiz:' . $this->moduleVersion,
            'test_mode' => ($this->testMode ? '1' : '0'),
        ];
    }

    /**
     * GET to Kiwiz API + decode response to JSON
     *
     * @param string $url
     * @param array $params
     * @throws \Kiwiz\Exceptions\ApiErrorException
     * @throws \Kiwiz\Exceptions\MalformedJsonException
     * @return json
     */
    public function get($url, $params = [])
    {
        // Construct parameters
        $params = array_merge($this->getDefaultParameters(), $params);
        try {
            $response = $this->client->get($url, [
                'query' => $params,
                'verify' => false
            ]);
            $json = json_decode($response->getBody());

            // Error when trying to decode JSON
            if (json_last_error() != JSON_ERROR_NONE || (isset($json->error) && $json->error === "Token")) {
                throw new MalformedJsonException('Unable to decode JSON');
            }

            return $json;
        } catch (RequestException $exception) {
            // Use dedicated method to handle this exception
            $this->handleRequestException($exception);
        }
    }

    /**
     * GET to Kiwiz API without JSON decode
     *
     * @param string $url
     * @param array $params
     * @throws \Kiwiz\Exceptions\ApiErrorException
     * @return string
     */
    public function getRaw($url, $params = [])
    {
        // Construct parameters
        $params = array_merge($this->getDefaultParameters(), $params);

        try {
            // Make the GET call
            $response = $this->client->get($url, [
                'query' => $params,
            ]);

            return (string)$response->getBody();
        } catch (RequestException $exception) {
            // Use dedicated method to handle this exception
            $this->handleRequestException($exception);
        }
    }

    /**
     * POST to Kiwiz API
     *
     * @param string $url
     * @param array $params
     * @throws \Kiwiz\Exceptions\ApiErrorException
     * @throws \Kiwiz\Exceptions\MalformedJsonException
     * @return json
     */
    public function post($url, $params = [])
    {
        try {
            // Make the POST call
            $response = $this->client->post($url, [
                'verify' => false,
                'body' => $params,
                'query' => $this->getDefaultParameters(),
            ]);
            $json = json_decode($response->getBody());

            // Error when trying to decode JSON
            if (json_last_error() != JSON_ERROR_NONE) {
                throw new MalformedJsonException('Unable to decode JSON');
            }

            return $json;
        } catch (RequestException $exception) {
            // Use dedicated method to handle this exception
            $this->handleRequestException($exception);
        }
    }

    /**
     * Handle RequestException, try to translate the error message to an explicit ApiErrorException
     *
     * @param \GuzzleHttp\Exception\RequestException $exception
     * @throws \Kiwiz\Exceptions\ApiErrorException
     * @return void
     */
    protected function handleRequestException(RequestException $exception)
    {
        // Return error returned by Kiwiz
        $response = $exception->getResponse();
        if (empty($response)) {
            // cURL error ?
            throw new ApiErrorException(sprintf('Error from API - ' . $exception->getMessage()));
        }

        // Try to decode response body to JSON
        $json = json_decode($response->getBody());

        // Error when trying to decode JSON
        if (json_last_error() != JSON_ERROR_NONE) {
            throw new ApiErrorException(sprintf('Error from API with unmanaged message: %s - HTTP Code: %d', $response->getBody(), $response->getStatusCode()), $response->getStatusCode());
        }

        // Retrieve error message from Kiwiz
        if (!empty($json->error) && !empty($json->message)) {
            throw new ApiErrorException(sprintf('Error from API - Message: %s - Error: %s - HTTP Code: %d', $json->message, $json->error, $response->getStatusCode()), $response->getStatusCode());
        } else {
            throw new ApiErrorException(sprintf('Error from API with unmanaged JSON message: %s - HTTP Code: %d', json_encode($json), $response->getStatusCode()), $response->getStatusCode());
        }
    }

    /**
     * Generate a token for authentication purposes
     *
     * @param string $username
     * @param string $password
     * @param int $subscriptionId
     * @return string|null
     */
    public function generateToken($username, $password, $subscriptionId)
    {
        $result = $this->post('/token/generate', [
            'username' => $username,
            'password' => $password,
            'subscription_id' => $subscriptionId,
        ]);

        if (!empty($result->token)) {
            return $result->token;
        }
    }

    /**
     * Retrieve current Kiwiz quota & add calculated consumption
     *
     * @return json
     */
    public function getQuota()
    {
        $quota = $this->get('/quota/info', ['verify' => false]);

        // Calculate current consumption in percentage
        $quota->consumption = isset($quota->limit) && $quota->limit ? round(($quota->used / $quota->limit) * 100) : 0;

        return $quota;
    }

    /**
     * Retrieve current Kiwiz plan
     *
     * @return json
     */
    public function getPlan()
    {
        return $this->get('/plan/get');
    }

    /**
     * Send PrestaShop invoice to Kiwiz API and return JSON response
     *
     * @param \Order $order
     * @return json
     */
    public function saveInvoice(\Order $order)
    {
        return $this->post('/invoice/save', [
            new PostFile('document', $this->pdfRetriever->order($order), $order->invoice_number . '.pdf'),
            'data' => json_encode($this->objectMapper->order($order)),
        ]);
    }

    /**
     * Retrieve invoice from Kiwiz API
     *
     * @param string $blockHash
     * @return string
     */
    public function getInvoice($blockHash)
    {
        return $this->getRaw('/invoice/get', [
            'block_hash' => $blockHash,
        ]);
    }

    /**
     * Send PrestaShop order slip to Kiwiz API and return JSON response
     *
     * @param \OrderSlip $orderSlip
     * @return json
     */
    public function saveCreditMemo(\OrderSlip $orderSlip)
    {

        // hmzeh
//       dump($this->objectMapper->orderSlip($orderSlip));
//       die;

        return $this->post('/creditmemo/save', [
            new PostFile('document', $this->pdfRetriever->orderSlip($orderSlip), $orderSlip->id . '.pdf'),
            'data' => json_encode($this->objectMapper->orderSlip($orderSlip)),
        ]);
    }

    /**
     * Retrieve order slip from Kiwiz API
     *
     * @param string $blockHash
     * @return string
     */
    public function getCreditMemo($blockHash)
    {
        return $this->getRaw('/creditmemo/get', [
            'block_hash' => $blockHash,
        ]);
    }

    /**
     * Retrieve a document from Kiwiz API
     *
     * @param \Kiwiz\Sync $doc
     * @return string
     */
    public function getDocument(Sync $doc)
    {
        if ($doc->document_type == 'order_invoice') {
            return $this->getInvoice($doc->block_hash);
        } elseif ($doc->document_type == 'order_slip') {
            return $this->getCreditMemo($doc->block_hash);
        }
    }
}
