<?php

namespace Kiwiz;

use \Tax;
use \Order;
use \Tools;
use \Address;
use \Carrier;
use \Country;
use \Product;
use \Customer;
use \Validate;
use \Exception;
use \OrderSlip;
use \OrderDetail;
use \Manufacturer;
use \Configuration;
use \TaxCalculator;
use \TaxManagerFactory;
if (!defined('_PS_VERSION_')) { exit; }
class ObjectMapper
{
    /**
     * List of current shop taxes
     *
     * @var array
     */
    private $taxes = [];

    /**
     * Init some attributes to improve performances
     *
     * @return void
     */
    public function __construct()
    {
        $this->idLang = (int)Configuration::get('PS_LANG_DEFAULT');
        $this->ecoTaxRulesGroupId = (int)Configuration::get('PS_ECOTAX_TAX_RULES_GROUP_ID');
        foreach (Tax::getTaxes($this->idLang, false) as $tax) {
            $this->taxes[(int)$tax['id_tax']] = $tax;
        }
        $this->taxAddressType = Configuration::get('PS_TAX_ADDRESS_TYPE');
    }

    /**
     * Merge all taxes, group by tax name
     *
     * @param array $taxesTable
     * @return array
     */
    protected function mergeTaxTable($taxesTable)
    {
        $newTaxesTable = array();
        foreach ($taxesTable as $taxesRow) {
            if (!isset($newTaxesTable[$taxesRow['tax_name']])) {
                $newTaxesTable[$taxesRow['tax_name']] = $taxesRow;
            } else {
                $newTaxesTable[$taxesRow['tax_name']]['tax_value'] += Tools::ps_round($taxesRow['tax_value'], 4);
            }
        }

        return array_values($newTaxesTable);
    }

    /**
     * Map a PrestaShop Order object to Kiwiz format
     *
     * @param \Order $order
     * @return array
     */
    public function order(Order $order)
    {
        // Check if Order object is valid
        if (!Validate::isLoadedObject($order)) {
            throw new Exception('Order object is not valid');
        }

        // Check if the current Order has an invoice
        if (!$order->hasInvoice()) {
            throw new Exception(sprintf('Order %d does not have invoice yet', $order->id));
        }

        $invoice = [];

        // Address used for tax calculation
        $addressTaxCalculation = new Address((int)$order->{$this->taxAddressType});
        // Check if Address object is valid
        if (!Validate::isLoadedObject($addressTaxCalculation)) {
            throw new Exception(sprintf('Unable to retrieve customer address with ID %d', (int)$order->{$this->taxAddressType}));
        }

        // Retrieve Customer
        $customer = new Customer($order->id_customer);
        // Check if Customer object is valid
        if (!Validate::isLoadedObject($customer)) {
            throw new Exception(sprintf('Unable to retrieve customer with ID %d', (int)$order->id_customer));
        }

        // Retrieve Carrier
        $carrier = new Carrier($order->id_carrier);
        // Check if Carrier object is valid
        if (!empty($order->id_carrier) && !Validate::isLoadedObject($carrier)) {
            throw new Exception(sprintf('Unable to retrieve carrier with ID %d', (int)$order->id_carrier));
        }

        // Retrieve EcoTax calculator
        $ecotaxCalculator = $this->getEcotaxCalculator($addressTaxCalculation);

        $invoice['increment_id'] = (int)$order->getInvoiceNumber($order->invoice_number);
        $invoice['date'] = $order->date_add;
        $invoice['email'] = $customer->email;
        $invoice['billing_address'] = $this->address(new Address($order->id_address_invoice));
        // Use payment if available, else use payment module name
        $invoice['payment_method'] = (!empty($order->payment) ? $order->payment : $order->module);
        $invoice['shipping_address'] = $this->address(new Address($order->id_address_delivery));
        $invoice['shipping_method'] = $carrier->name;
        $invoice['shipping_amount_excl_tax'] = Tools::ps_round((float)$order->total_shipping_tax_excl, 4);
        $invoice['shipping_tax_amount'] = [];
        if (!empty($carrier)) {
            // Retrieve Tax calculator
            $taxCalculator = $carrier->getTaxCalculator($addressTaxCalculation);
            $invoice['shipping_tax_amount'] = $this->taxes($taxCalculator, (float)$order->total_shipping_tax_excl);
        }
        $invoice['items'] = $this->items($addressTaxCalculation, $order);
        $invoice['grand_total_excl_tax'] = Tools::ps_round((float)$order->total_paid_tax_excl, 4);
        $invoice['grand_total_tax_amount'] = $this->orderTaxes($order);

        // Add carrier tax to taxes list
        foreach ($invoice['shipping_tax_amount'] as $carrierTaxTable) {
            $invoice['grand_total_tax_amount'][] = $carrierTaxTable;
        }

        // Add ecotax to taxes list
        if ($ecotaxCalculator instanceof TaxCalculator && $this->getTotalEcoTax($order)) {
            $invoice['grand_total_tax_amount'][] = $this->orderEcoTaxes($ecotaxCalculator, $order);
        }

        // Merge all taxes
        $invoice['grand_total_tax_amount'] = $this->mergeTaxTable($invoice['grand_total_tax_amount']);

        return $invoice;
    }

    /**
     * Map a PrestaShop OrderSlip object to Kiwiz format
     *
     * @param \OrderSlip $orderSlip
     * @return array
     */
    public function orderSlip(OrderSlip $orderSlip)
    {
        $creditMemo = [];

        // Retrieve Customer
        $customer = new Customer($orderSlip->id_customer);
        // Check if Customer object is valid
        if (!Validate::isLoadedObject($customer)) {
            throw new Exception(sprintf('Unable to retrieve customer with ID %d', (int)$orderSlip->id_customer));
        }

        // Retrieve Order
        $order = new Order($orderSlip->id_order);
        // Check if Order object is valid
        if (!Validate::isLoadedObject($order)) {
            throw new Exception(sprintf('Unable to retrieve Order with ID %d', (int)$orderSlip->id_order));
        }

        // Address used for tax calculation
        $addressTaxCalculation = new Address((int)$order->{$this->taxAddressType});
        // Check if Address object is valid
        if (!Validate::isLoadedObject($addressTaxCalculation)) {
            throw new Exception(sprintf('Unable to retrieve customer address with ID %d', (int)$order->{$this->taxAddressType}));
        }

        // Retrieve Carrier
        $carrier = new Carrier($order->id_carrier);
        // Check if Carrier object is valid
        if (!empty($order->id_carrier) && !Validate::isLoadedObject($carrier)) {
            throw new Exception(sprintf('Unable to retrieve carrier with ID %d', (int)$order->id_carrier));
        }

        // Retrieve EcoTax calculator
        $ecotaxCalculator = $this->getEcotaxCalculator($addressTaxCalculation);

        $creditMemo['increment_id'] = $orderSlip->id;
        $creditMemo['date'] = $orderSlip->date_add;
        $creditMemo['grand_total_excl_tax'] = Tools::ps_round(((float)$orderSlip->total_products_tax_excl + (float)$orderSlip->total_shipping_tax_excl), 4);
        $creditMemo['grand_total_tax_amount'] = $this->orderSlipTaxes($orderSlip, $order);
        $creditMemo['email'] = $customer->email;

        if (!empty($carrier)) {
            // Retrieve Tax calculator
            $taxCalculator = $carrier->getTaxCalculator($addressTaxCalculation);
            // Add carrier tax to taxes list
            foreach ($this->taxes($taxCalculator, (float)$orderSlip->total_shipping_tax_excl) as $carrierTaxTable) {
                $creditMemo['grand_total_tax_amount'][] = $carrierTaxTable;
            }
        }

        // Add ecotax to taxes list
        if ($ecotaxCalculator instanceof TaxCalculator && $this->getTotalEcoTax($order)) {
            $creditMemo['grand_total_tax_amount'][] = $this->orderEcoTaxes($ecotaxCalculator, $order);
        }

        // Merge all taxes
        $creditMemo['grand_total_tax_amount'] = $this->mergeTaxTable($creditMemo['grand_total_tax_amount']);

        return $creditMemo;
    }

    /**
     * Calculate the total amount of eco-tax for a specific order
     *
     * @param \Order $order
     * @return float
     */
    protected function getTotalEcoTax(Order $order)
    {
        $totalEcoTax = 0;

        foreach ($order->getOrderDetailList() as $orderItem) {
            // Eco-tax
            if ($orderItem['ecotax'] > 0) {
                $totalEcoTax += (float)$orderItem['ecotax'];
            }
        }

        return $totalEcoTax;
    }

    /**
     * Map total ecotaxe to Kiwiz format
     *
     * @param \TaxCalculator $taxCalculator
     * @param float $amount
     * @return array
     */
    protected function ecoTaxes(TaxCalculator $taxCalculator, $amount)
    {
        return [
            'tax_name' => 'Ecotax - ' . $taxCalculator->getTaxesName(),
            'tax_value' => Tools::ps_round(max(0, $amount), 4),
        ];
    }

    /**
     * Map total Order taxes to Kiwiz format
     *
     * @param array $orderDetails
     * @param array $orderDetailsTaxes
     * @return array
     */
    protected function calculateOrderTaxes(array $orderDetails, array $orderDetailsTaxes)
    {
        $taxes = [];
        $totalTaxesAmount = [];

        foreach ($orderDetails as $orderItem) {
            $itemId = (int)$orderItem['id_order_detail'];
            $orderItemTaxes = null;
            // Retrieve order item row tax (if exists)
            foreach ($orderDetailsTaxes as $orderItemTaxesRow) {
                if ($orderItemTaxesRow['id_order_detail'] == $itemId) {
                    $orderItemTaxes = $orderItemTaxesRow;
                    break;
                }
            }
            // Apply price provided by Order::getProductTaxesDetails() (include reduction)
            if (!empty($orderItemTaxes)) {
                $orderItem['total_price_tax_excl'] = $orderItemTaxesRow['total_tax_base'];
            }

            // Retrieve the right tax calculator depending on the current order detail item
            $taxCalculator = OrderDetail::getTaxCalculatorStatic($itemId);
            $taxesAmount = $taxCalculator->getTaxesAmount((float)$orderItem['total_price_tax_excl']);

            // Sum for each tax group
            foreach ($taxesAmount as $idTaxGroup => $amount) {
                if (!isset($totalTaxesAmount[$idTaxGroup])) {
                    $totalTaxesAmount[$idTaxGroup] = 0;
                }
                $totalTaxesAmount[$idTaxGroup] += $amount;
            }
        }

        // Format all taxes to Kiwiz format
        foreach ($totalTaxesAmount as $idTaxGroup => $taxAmount) {
            $taxes[] = [
                'tax_name' => $this->taxes[$idTaxGroup]['name'],
                'tax_value' => Tools::ps_round(max(0, $taxAmount), 4),
            ];
        }

        return $taxes;
    }

    /**
     * Map total Order taxes to Kiwiz format
     *
     * @param \Order $order
     * @return array
     */
    protected function orderTaxes(Order $order)
    {
        return $this->calculateOrderTaxes($order->getOrderDetailList(), $order->getProductTaxesDetails());
    }

    /**
     * Map total OrderSlip taxes to Kiwiz format
     *
     * @param \OrderSlip $orderSlip
     * @param \Order $order
     * @return array
     */
    protected function orderSlipTaxes(OrderSlip $orderSlip, Order $order)
    {
        return $this->calculateOrderTaxes(OrderSlip::getOrdersSlipProducts($orderSlip->id, $order), array());
    }

    /**
     * Map total Order ecotaxe to Kiwiz format
     *
     * @param \TaxCalculator $taxCalculator
     * @param \Order $order
     * @return array
     */
    protected function orderEcoTaxes(TaxCalculator $taxCalculator, Order $order)
    {
        $totalEcoTax = $this->getTotalEcoTax($order);

        return $this->ecoTaxes($taxCalculator, $totalEcoTax);
    }

    /**
     * Retrieve ecotax calculator
     *
     * @param \Address $addressTaxCalculation
     * @return \TaxCalculator|null
     */
    protected function getEcotaxCalculator(Address $addressTaxCalculation)
    {
        // Retrieve EcoTax calculator
        $ecotaxCalculator = null;
        if ($this->ecoTaxRulesGroupId) {
            $ecotaxManager = TaxManagerFactory::getManager($addressTaxCalculation, $this->ecoTaxRulesGroupId);
            $ecotaxCalculator = $ecotaxManager->getTaxCalculator();
        } else {
            $ecotaxCalculator = new TaxCalculator(array());
        }

        return $ecotaxCalculator;
    }

    /**
     * Map a PrestaShop Address object to Kiwiz format
     *
     * @param \Address $address
     * @return array
     */
    public function address(Address $address)
    {
        $address = [
            'firstname' => $address->firstname,
            'lastname' => $address->lastname,
            'company' => $address->company,
            'street' => $address->address1,
            'postcode' => $address->postcode,
            'city' => $address->city,
            'country_code' => Country::getIsoById($address->id_country),
        ];

        return $address;
    }

    /**
     * Map taxes to Kiwiz format
     *
     * @param \TaxCalculator $taxCalculator
     * @param float $amount
     * @return array
     */
    protected function taxes(TaxCalculator $taxCalculator, $amount)
    {
        $taxes = [];

        $taxValue = $taxCalculator->getTaxesTotalAmount($amount);
        if ($taxValue == 0) {
            return $taxes;
        }

        $taxes[] = [
            'tax_name' => $taxCalculator->getTaxesName(),
            'tax_value' => Tools::ps_round(max(0, $taxValue), 4),
        ];

        return $taxes;
    }

    /**
     * Map product list to Kiwiz format
     *
     * @param \Address $addressTaxCalculation
     * @param \Order $order
     * @return array
     */
    protected function items(Address $addressTaxCalculation, Order $order)
    {
        // Retrieve EcoTax calculator
        $ecotaxCalculator = $this->getEcotaxCalculator($addressTaxCalculation);
        $orderDetailsTaxes = $order->getProductTaxesDetails();

        $items = [];
        // Build items list
        foreach ($order->getOrderDetailList() as $orderItem) {
            $itemId = (int)$orderItem['id_order_detail'];
            $orderItemTaxes = null;
            // Retrieve order item row tax (if exists)
            foreach ($orderDetailsTaxes as $orderItemTaxesRow) {
                if ($orderItemTaxesRow['id_order_detail'] == $itemId) {
                    $orderItemTaxes = $orderItemTaxesRow;
                    break;
                }
            }
            // Apply price provided by Order::getProductTaxesDetails (include reduction)
            if (!empty($orderItemTaxes)) {
                $orderItem['total_price_tax_excl'] = $orderItemTaxesRow['total_tax_base'];
            }

            $items[$itemId] = [
                'sku' => $orderItem['product_reference'],
                'ean13' => $orderItem['product_ean13'],
                'product_name' => $orderItem['product_name'],
                'qty' => (int)$orderItem['product_quantity'],
                'row_total_excl_tax' => Tools::ps_round(((float)$orderItem['total_price_tax_excl']), 4),
                'row_total_tax_amount' => $this->taxes(OrderDetail::getTaxCalculatorStatic($itemId), (float)$orderItem['total_price_tax_excl']),
            ];
            // Eco-tax
            if ($orderItem['ecotax'] > 0) {
                $items[$itemId]['row_total_tax_amount'][] = $this->ecoTaxes($ecotaxCalculator, (float)$orderItem['ecotax']);
            }
            $product = new Product($orderItem['product_id']);
            if (Validate::isLoadedObject($product)) {
                // Retrieve manufacturer
                $manufacturerName = Manufacturer::getNameById($product->id_manufacturer);
                if (!empty($manufacturerName)) {
                    $items[$itemId]['manufacturer'] = $manufacturerName;
                }
            }
        }

        return $items;
    }
}
