<?php

namespace Kiwiz;

use \PDF;
use \Order;
use \Module;
use \Validate;
use \OrderSlip;
use \OrderInvoice;
use \PDFGenerator;
use \Configuration;
use Kiwiz\Api\Client;
use setasign\Fpdi\Tcpdf\Fpdi;
use setasign\Fpdi\PdfParser\StreamReader;
if (!defined('_PS_VERSION_')) { exit; }
class PdfRenderer
{
    /**
     * Constructor
     *
     * @param \Kiwiz\Api\Client $apiClient
     * @param \PDF $pdfObject
     * @param boolean $display
     * @return void
     */
    public function __construct(Client $apiClient, PDF $pdfObject, $display = true)
    {
        $this->pdfRetriever = new PdfRetriever();
        $this->apiClient = $apiClient;
        $this->pdfObject = $pdfObject;
        $this->display = $display;
        $this->module = Module::getInstanceByName('kiwiz');
    }

    /**
     * Use the native PDF render process
     *
     * @return mixed
     */
    private function nativePdfRender()
    {
        // Use the native process to render the PDF file
        $this->pdfObject->useClassicProcess = true;

        return $this->pdfObject->render($this->display);
    }

    /**
     * Retrieve Order or OrderSlip object, depending on the original object
     *
     * @param \Order|\OrderInvoice|\OrderSlip $object
     * @return \Order|\OrderSlip|null
     */
    private function retrieveObject($object)
    {
        if ($this->pdfObject->template == 'Invoice' && $object instanceof OrderInvoice) {
            return (new Order($object->id_order));
        } elseif ($this->pdfObject->template == 'Invoice' && $object instanceof Order) {
            return $object;
        } elseif ($this->pdfObject->template == 'OrderSlip' && $object instanceof OrderSlip) {
            return $object;
        }

        return null;
    }

    /**
     * Determinate PDF filename
     *
     * @return string
     */
    private function getFilename()
    {
        // Detect template using first object
        $pdfTemplate = $this->pdfObject->getTemplateObject($this->pdfObject->objects[0]);

        // Set filename
        $filename = $pdfTemplate->getFilename();
        if (count($this->pdfObject->objects) > 1) {
            $filename = $pdfTemplate->getBulkFilename();
        }
        $filename = ltrim($filename, '#');

        return $filename;
    }

    /**
     * Render a PDF from a single object
     *
     * @return string
     */
    public function renderPdf()
    {
        $obj = null;

        // Retrieve Order / OrderSlip
        $obj = $this->retrieveObject($this->pdfObject->objects[0]);
        if ($obj === null || !Validate::isLoadedObject($obj)) {
            // Use the native process in case we aren't able to find original order / order slip
            return $this->nativePdfRender();
        }

        // Check document date_add & compare it with Kiwiz activation date
        $activationDate = Configuration::get('KIWIZ_ACTIVATION_DATE');
        if ($obj instanceof Order) {
            // Check order date or invoice date
            if ($this->module->getOrderOrInvoiceDate($obj) < $activationDate) {
                // Use the native process
                return $this->nativePdfRender();
            }
        } elseif (strtotime($obj->date_add) < $activationDate) {
            // Use the native process
            return $this->nativePdfRender();
        }

        $document = Sync::findByObject($obj);
        // Try to retrieve the existing document
        if (!Validate::isLoadedObject($document)) {
            // Use the native process
            return $this->nativePdfRender();
        }

        // Return the PDF content from Kiwiz API
        try {
            return $this->apiClient->getDocument($document);
        } catch (\Exception $e) {
            // Return a new PDF page with a Kiwiz error
            $pdfGenerator = new PDFGenerator();
            $pdfGenerator->AddPage('P', 'A4');
            $pdfGenerator->writeHTML(sprintf($this->module->l('Error while retrieving the document (ID: %d - Type: %s) on Kiwiz.'), $document->id_document, $document->document_type));
            return $pdfGenerator->Output('error-kiwiz.pdf', 'S');
        }
    }

    /**
     * Render a PDF from a multiple objects
     *
     * @return string
     */
    public function renderMultiplePdf()
    {
        $fpdi = new Fpdi();
        $activationDate = Configuration::get('KIWIZ_ACTIVATION_DATE');

        // Loop on each Order/OrderSlip that has to be added to the PDF file
        foreach ($this->pdfObject->objects as $object) {
            // Retrieve Order / OrderSlip
            $obj = $this->retrieveObject($object);
            if ($obj === null || !Validate::isLoadedObject($obj)) {
                continue;
            }

            // Check document date_add & compare it with Kiwiz activation date
            if (strtotime($obj->date_add) < $activationDate) {
                // Retrieve original document file (generated by PrestaShop) as it was created prior Kiwiz activation
                $pdf = $this->pdfRetriever->getPdf($obj);
            } else {
                // Try to retrieve the existing document
                $document = Sync::findByObject($obj);
                if (!Validate::isLoadedObject($document)) {
                    // Retrieve original PDF file (generated by PrestaShop)
                    $pdf = $this->pdfRetriever->getPdf($obj);
                } else {
                    // Get PDF content from Kiwiz API
                    try {
                        $pdf = $this->apiClient->getDocument($document);
                    } catch (\Exception $e) {
                        // Return a new PDF page with a Kiwiz error
                        $pdfGenerator = new PDFGenerator();
                        $pdfGenerator->AddPage('P', 'A4');
                        $pdfGenerator->writeHTML(sprintf($this->module->l('Error while retrieving the document (ID: %d - Type: %s) on Kiwiz.'), $document->id_document, $document->document_type));
                        $pdf = $pdfGenerator->Output('error-kiwiz.pdf', 'S');
                    }
                }
            }

            // Concatenate the document into the PDF
            // Retrieve the page count
            $pageCount = $fpdi->setSourceFile(StreamReader::createByString($pdf));
            // Iterate through all pages and add each page to the PDF file
            for ($pageNo = 1; $pageNo <= $pageCount; $pageNo++) {
                // Import the current page
                $templateId = $fpdi->importPage($pageNo);
                // Retrieve the size of the page we want to import
                $pageSize = $fpdi->getTemplateSize($templateId);
                // Prevent TCPDF to add header & footer twice as we only want to concatenate our files
                $fpdi->SetPrintHeader(false);
                $fpdi->SetPrintFooter(false);
                // Add the page using its original orientation and size
                $fpdi->AddPage($pageSize['orientation'], $pageSize);
                $fpdi->useTemplate($templateId);
            }
        }

        // Return the PDF content
        return $fpdi->Output($this->getFilename(), 'S');
    }

    /**
     * Dispatch on the right render method, depending on the number of objects to render
     * Output the PDF content to the buffer or make force the browser to download the file
     *
     * @return mixed
     */
    public function render()
    {
        // This method should only be called for specific PDF template
        if (!in_array($this->pdfObject->template, array('Invoice', 'OrderSlip'))) {
            return $this->nativePdfRender();
        }
        // Define a flag on the PDF object in order to force the native render process (if true)
        $this->pdfObject->useClassicProcess = false;

        // Use the right render method, depending on the objects count
        if (count($this->pdfObject->objects) > 1) {
            $pdfContent = $this->renderMultiplePdf();
        } else {
            $pdfContent = $this->renderPdf();
        }

        // Force web browser to download the PDF
        if ($this->display) {
            $filename = $this->getFilename();

            // From TCPDF
            header('Content-Type: application/pdf');
            header('Cache-Control: private, must-revalidate, post-check=0, pre-check=0, max-age=1');
            header('Pragma: public');
            header('Expires: Sat, 26 Jul 1997 05:00:00 GMT');
            header('Last-Modified: '.gmdate('D, d M Y H:i:s').' GMT');
            header('Content-Disposition: attachment; filename="'.basename($filename).'"');
            header('Content-Transfer-Encoding: binary');

            die($pdfContent);
        }

        // Return the PDF content (used for emails etc.)
        return $pdfContent;
    }
}
