<?php

namespace Kiwiz;

use \PDF;
use \Hook;
use \Order;
use \Tools;
use \Context;
use \OrderSlip;
use \ObjectModel;
if (!defined('_PS_VERSION_')) { exit; }
class PdfRetriever
{
    /**
     * Reset font to be used into PDF
     * Prevent TTF file to be embedded
     *
     * @param \PDF $pdf
     * @return void
     */
    protected function removeFonts(PDF $pdf)
    {
        $currentIsoCode = Context::getContext()->language->iso_code;
        // Reset font, prevent TTF to be embedded
        $pdf->pdf_renderer->font = '';
        $pdf->pdf_renderer->font_by_lang[$currentIsoCode] = '';
    }

    /**
     * Retrieve invoice from PrestaShop
     *
     * @param \Order $order
     * @return string
     */
    public function order(Order $order)
    {
        $orderInvoiceList = $order->getInvoicesCollection();
        Hook::exec('actionPDFInvoiceRender', [
            'order_invoice_list' => $orderInvoiceList,
        ]);

        $pdf = new PDF($orderInvoiceList, PDF::TEMPLATE_INVOICE, Context::getContext()->smarty);
        $pdf->useClassicProcess = true;
        $pdf->pdf_renderer->setDocCreationTimestamp(strtotime($order->date_add));
        $pdf->pdf_renderer->setDocModificationTimestamp(strtotime($order->date_add));

        // Remove embedded fonts from the PDF
        if (version_compare(_PS_VERSION_, '1.7.0.0', '<')) {
            $this->removeFonts($pdf);
        }

        return $this->setDocumentUuid($pdf->render(false), 'invoice-' . (int)$order->id . $order->date_add);
    }

    /**
     * Retrieve credit memo from PrestaShop
     *
     * @param \OrderSlip $orderSlip
     * @return string
     */
    public function orderSlip(OrderSlip $orderSlip)
    {
        $pdf = new PDF($orderSlip, PDF::TEMPLATE_ORDER_SLIP, Context::getContext()->smarty);
        $pdf->useClassicProcess = true;
        $pdf->pdf_renderer->setDocCreationTimestamp(strtotime($orderSlip->date_add));
        $pdf->pdf_renderer->setDocModificationTimestamp(strtotime($orderSlip->date_add));

        // Remove embedded fonts from the PDF
        if (version_compare(_PS_VERSION_, '1.7.0.0', '<')) {
            $this->removeFonts($pdf);
        }

        return $this->setDocumentUuid($pdf->render(false), 'creditmemo-' . (int)$orderSlip->id . $orderSlip->date_add);
    }

    /**
     * Retrieve PDF content depending of object class
     *
     * @param \Order|\OrderSlip $object
     * @return string
     */
    public function getPdf(ObjectModel $object)
    {
        if ($object instanceof Order) {
            return $this->order($object);
        } elseif ($object instanceof OrderSlip) {
            return $this->orderSlip($object);
        }
    }

    /**
     * Find the UUID into the PDF and replace it using the new one
     *
     * @param string $pdfContent
     * @param string $newUuid
     * @return string
     */
    private function setDocumentUuid($pdfContent, $newUuid)
    {
        // Look for Document/Instance UUID into the PDF content
        if (preg_match('#<xmpMM:(DocumentID|InstanceID)>uuid:([0-9a-f-]+)<\/xmpMM:(DocumentID|InstanceID)>#mi', $pdfContent, $matches)) {
            $uuid = $matches[2];
            // Make a MD5 hash of the new UUID (as TCPDF do)
            $newFileId = md5($newUuid);
            // Create new UUID from MD5 (as TCPDF do)
            $newUuid = Tools::substr($newFileId, 0, 8) . '-' . Tools::substr($newFileId, 8, 4) . '-' . Tools::substr($newFileId, 12, 4) . '-' . Tools::substr($newFileId, 16, 4) . '-' . Tools::substr($newFileId, 20, 12);

            // Replace UUID that was generated by TCPDF by a static one (per invoice/creditmemo)
            $pdfContent = str_replace(str_replace('-', '', $uuid), $newFileId, $pdfContent);
            $pdfContent = str_replace($uuid, $newUuid, $pdfContent);
        }

        return $pdfContent;
    }
}
