<?php

namespace Kiwiz;

use \Db;
use \Order;
use \Context;
use \Validate;
use \OrderSlip;
use \ObjectModel;
use \Configuration;
use Kiwiz\Exceptions\Exception;
if (!defined('_PS_VERSION_')) { exit; }
class Sync extends ObjectModel
{
    /**
     * ID of the document (id_order_invoice / id_order_slip)
     *
     * @var int
     */
    public $id_document;

    /**
     * Type of document (order_invoice, order_slip)
     *
     * @var string
     */
    public $document_type;

    /**
     * Corresponding document hash on Kiwiz
     *
     * @var string
     */
    public $file_hash;

    /**
     * Corresponding hash block on Kiwiz
     *
     * @var string
     */
    public $block_hash;

    /**
     * When the document has been sent to Kiwiz
     *
     * @var string
     */
    public $date_add;

    /**
     * @see ObjectModel::$definition
     */
    public static $definition = array(
        'table' => 'kiwiz_sync',
        'primary' => 'id_kiwiz_sync',
        'fields' => array(
            'id_document'   => array('type' => self::TYPE_INT, 'validate' => 'isUnsignedId', 'required' => true),
            'document_type' => array('type' => self::TYPE_STRING, 'required' => true),
            'file_hash'     => array('type' => self::TYPE_STRING, 'required' => true),
            'block_hash'    => array('type' => self::TYPE_STRING, 'required' => true),
            'date_add'      => array('type' => self::TYPE_DATE, 'required' => true),
        ),
    );

    /**
     * Retrieve current document for the order
     *
     * @param \Order|\OrderSlip $obj
     * @return \Kiwiz\Sync|null
     */
    public static function findByObject(ObjectModel $obj)
    {
        if ($obj instanceof Order) {
            $documentType = 'order_invoice';
        } elseif ($obj instanceof OrderSlip) {
            $documentType = 'order_slip';
        } else {
            throw new Exception(sprintf('Sync::findByObject - Unmanaged ObjectModel %s', get_class($obj)));
        }

        // Find the document id into the dedicated table
        $idDocument = Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue(
            'SELECT `id_kiwiz_sync`
            FROM `'._DB_PREFIX_.'kiwiz_sync`
            WHERE `id_document`=' . (int)$obj->id . '
            AND `document_type`="' . pSQL($documentType) . '"'
        );

        // Load the object
        if (!empty($idDocument)) {
            return (new self($idDocument));
        }

        return null;
    }

    /**
     * Retrieve current existing document for the order, or create a new one
     *
     * @param \Order $order
     * @param \stdClass $kiwizData
     * @return \Kiwiz\Sync
     */
    public static function findOrCreateInvoice(Order $order, \stdClass $kiwizData)
    {
        // Try to retrieve the existing document
        $document = self::findByObject($order);
        if (Validate::isLoadedObject($document)) {
            return $document;
        }

        // Create the document entry
        $obj = new static();
        $obj->id_document = (int)$order->id;
        $obj->document_type = 'order_invoice';
        $obj->file_hash = $kiwizData->file_hash;
        $obj->block_hash = $kiwizData->block_hash;

        // Save into the database
        if (!$obj->save()) {
            throw new Exception('Sync::findOrCreateInvoice - Unable to save ObjectModel');
        }

        return $obj;
    }

    /**
     * Retrieve current existing document for the order slip, or create a new one
     *
     * @param \OrderSlip $orderSlip
     * @param \stdClass $kiwizData
     * @return \Kiwiz\Sync
     */
    public static function findOrCreateCreditMemo(OrderSlip $orderSlip, \stdClass $kiwizData)
    {
        // Try to retrieve the existing document
        $document = self::findByObject($orderSlip);
        if (Validate::isLoadedObject($document)) {
            return $document;
        }

        // Create the document entry
        $obj = new static();
        $obj->id_document = (int)$orderSlip->id;
        $obj->document_type = 'order_slip';
        $obj->file_hash = $kiwizData->file_hash;
        $obj->block_hash = $kiwizData->block_hash;

        // Save into the database
        if (!$obj->save()) {
            throw new Exception('Sync::findOrCreateCreditMemo - Unable to save ObjectModel');
        }

        return $obj;
    }

    /**
     * Retrieve all orders that are not already synced to Kiwiz
     *
     * @param array $idList
     * @return array
     */
    public static function getOrdersToSync(array $idList = array())
    {
        $activationDate = date('Y-m-d H:i:s', Configuration::get('KIWIZ_ACTIVATION_DATE'));
        $idShop = Context::getContext()->shop->id;

        // Retrieve all orders with at least one invoice in database
        $idList = Db::getInstance(_PS_USE_SQL_SLAVE_)->ExecuteS(
            'SELECT o.`id_order`
            FROM `' . _DB_PREFIX_ . 'orders` o
            JOIN `' . _DB_PREFIX_ . 'order_invoice` oi ON (o.`id_order` = oi.`id_order`)
            WHERE o.`date_add` >= "' . pSQL($activationDate) . '"
            ' . (!empty($idList) ? 'AND o.`id_order` IN (' . implode(',', array_map('intval', $idList)) . ')' : '') . '
            AND o.`id_order` NOT IN (
                SELECT `id_document`
                FROM `' . _DB_PREFIX_ . 'kiwiz_sync`
                WHERE `document_type`="order_invoice"
            )'
        );

        if (!is_array($idList) || empty($idList)) {
            return array();
        }

        // Retrieve & return all ID
        return array_map(
            function ($row) {
                return (int)$row['id_order'];
            },
            $idList
        );
    }

    /**
     * Retrieve all order slip that are not already synced to Kiwiz
     *
     * @param array $idList
     * @return array
     */
    public static function getOrderSlipToSync(array $idList = array())
    {
        $activationDate = date('Y-m-d H:i:s', Configuration::get('KIWIZ_ACTIVATION_DATE'));

        // Retrieve all order slip in database
        $idList = Db::getInstance(_PS_USE_SQL_SLAVE_)->ExecuteS(
            'SELECT `id_order_slip`
            FROM `' . _DB_PREFIX_ . 'order_slip`
            WHERE `date_add` >= "' . pSQL($activationDate) . '"
            ' . (!empty($idList) ? 'AND `id_order_slip` IN (' . implode(',', array_map('intval', $idList)) . ')' : '') . '
            AND `id_order_slip` NOT IN (
                SELECT `id_document`
                FROM `' . _DB_PREFIX_ . 'kiwiz_sync`
                WHERE `document_type`="order_slip"
            )'
        );

        if (!is_array($idList) || empty($idList)) {
            return array();
        }

        // Retrieve & return all ID
        return array_map(
            function ($row) {
                return (int)$row['id_order_slip'];
            },
            $idList
        );
    }
}
